@@ -1,17 +1,18 @@
 import datetime
 from functools import partial
 import hashlib
+import io
 import os
 from pkg_resources import parse_version
 import re
 import warnings
+import weakref
 
 import numpy as np
 from astropy.io import fits
 from astropy import time
 from astropy.utils.exceptions import AstropyWarning
 import asdf
-from asdf import fits_embed
 from asdf import resolver
 from asdf import schema as asdf_schema
 from asdf.tags.core import NDArrayType
@@ -19,6 +20,7 @@ from asdf.tags.core import ndarray, HistoryEntry
 from asdf import treeutil
 from asdf.util import HashableDict
 from asdf import tagged
+from asdf import generic_io
 from jsonschema import validators
 
 from . import properties
@@ -34,6 +36,10 @@ log.addHandler(logging.NullHandler())
 __all__ = ['to_fits', 'from_fits', 'fits_hdu_name', 'get_hdu', 'is_builtin_fits_keyword']
 
 
+_ASDF_EXTENSION_NAME = "ASDF"
+_FITS_SOURCE_PREFIX = "fits:"
+_NDARRAY_TAG = "tag:stsci.edu:asdf/core/ndarray-1.0.0"
+
 _ASDF_GE_2_6 = parse_version(asdf.__version__) >= parse_version('2.6')
 
 
@@ -176,6 +182,17 @@ def _make_hdu(hdulist, hdu_name, index=None, hdu_type=None, value=None):
 
 
 def _get_or_make_hdu(hdulist, hdu_name, index=None, hdu_type=None, value=None):
+    if isinstance(hdulist, weakref.ReferenceType):
+        ref = hdulist()
+        result = _get_or_make_hdu(ref, hdu_name, index=index, hdu_type=hdu_type, value=value)
+        # While likely not needed (as ref will fall out of scope on
+        # return), del ref to remove the reference to the hdulist we
+        # resolved from the weakref. weakref is important here as
+        # this function is called within a validator which will be
+        # cached holding referenced objects in memory.
+        # https://github.com/spacetelescope/stdatamodels/pull/109
+        del ref
+        return result
     try:
         hdu = get_hdu(hdulist, hdu_name, index=index)
     except AttributeError:
@@ -312,7 +329,7 @@ def _fits_type(fits_context, validator, items, instance, schema):
 
 class FitsContext:
     def __init__(self, hdulist):
-        self.hdulist = hdulist
+        self.hdulist = weakref.ref(hdulist)
         self.comment_stack = []
         self.sequence_index = None
 
@@ -349,13 +366,16 @@ FITS_SCHEMA_URL_MAPPING = resolver.Resolver(
 
 
 def _save_from_schema(hdulist, tree, schema):
-    def convert_datetimes(node, json_id):
+    def datetime_callback(node, json_id):
         if isinstance(node, datetime.datetime):
             node = time.Time(node)
+
         if isinstance(node, time.Time):
             node = str(time.Time(node, format='iso'))
+
         return node
-    tree = treeutil.walk_and_modify(tree, convert_datetimes)
+
+    tree = treeutil.walk_and_modify(tree, datetime_callback)
 
     if _ASDF_GE_2_6:
         kwargs = {"_visit_repeat_nodes": True}
@@ -368,6 +388,48 @@ def _save_from_schema(hdulist, tree, schema):
     # This actually kicks off the saving
     validator.validate(tree, _schema=schema)
 
+    # Replace arrays in the tree that are identical to HDU arrays
+    # with ndarray-1.0.0 tagged objects with special source values
+    # that represent links to the surrounding FITS file.
+    def ndarray_callback(node, json_id):
+        if (isinstance(node, (np.ndarray, NDArrayType))):
+            for hdu_index, hdu in enumerate(hdulist):
+                if hdu.data is not None and node is hdu.data:
+                    return _create_tagged_dict_for_fits_array(hdu, hdu_index)
+
+        return node
+
+    tree = treeutil.walk_and_modify(tree, ndarray_callback)
+
+    return tree
+
+
+def _create_tagged_dict_for_fits_array(hdu, hdu_index):
+     # Views over arrays stored in FITS files have some idiosyncrasies.
+     # astropy.io.fits always writes arrays C-contiguous with big-endian
+     # byte order, whereas asdf preserves the "contiguousity" and byte order
+     # of the base array.
+    dtype, byteorder = ndarray.numpy_dtype_to_asdf_datatype(
+        hdu.data.dtype,
+        include_byteorder=True,
+        override_byteorder="big"
+    )
+
+    if hdu.name == "":
+        source = f"{_FITS_SOURCE_PREFIX}{hdu_index}"
+    else:
+        source = f"{_FITS_SOURCE_PREFIX}{hdu.name},{hdu.ver}"
+
+    return tagged.TaggedDict(
+        data={
+            "source": source,
+            "shape": list(hdu.data.shape),
+            "datatype": dtype,
+            "byteorder": byteorder
+        },
+        tag=_NDARRAY_TAG
+    )
+
 
 def _normalize_arrays(tree):
     """
@@ -430,14 +492,30 @@ def to_fits(tree, schema):
     hdulist.append(fits.PrimaryHDU())
 
     tree = _normalize_arrays(tree)
-    _save_from_schema(hdulist, tree, schema)
+    tree = _save_from_schema(hdulist, tree, schema)
     _save_extra_fits(hdulist, tree)
     _save_history(hdulist, tree)
 
     # Store the FITS hash in the tree
     tree[FITS_HASH_KEY] = fits_hash(hdulist)
 
-    return hdulist, tree
+    if _ASDF_EXTENSION_NAME in hdulist:
+        del hdulist[_ASDF_EXTENSION_NAME]
+
+    hdulist.append(_create_asdf_hdu(tree))
+
+    return hdulist
+
+
+def _create_asdf_hdu(tree):
+    buffer = io.BytesIO()
+    asdf.AsdfFile(tree).write_to(buffer)
+    buffer.seek(0)
+
+    data = np.array(buffer.getbuffer(), dtype=np.uint8)[None, :]
+    fmt = f"{len(data[0])}B"
+    column = fits.Column(array=data, format=fmt, name="ASDF_METADATA")
+    return fits.BinTableHDU.from_columns([column], name=_ASDF_EXTENSION_NAME)
 
 
 ##############################################################################
@@ -644,10 +722,52 @@ def from_fits_asdf(hdulist,
     Wrap asdf call to extract optional arguments
     """
     ignore_missing_extensions = kwargs.pop('ignore_missing_extensions')
-    return fits_embed.AsdfInFits.open(hdulist,
-                                      ignore_version_mismatch=ignore_version_mismatch,
-                                      ignore_unrecognized_tag=ignore_unrecognized_tag,
-                                      ignore_missing_extensions=ignore_missing_extensions)
+
+    try:
+        asdf_extension = hdulist[_ASDF_EXTENSION_NAME]
+    except (KeyError, IndexError, AttributeError):
+        # This means there is no ASDF extension
+        return asdf.AsdfFile(
+            ignore_version_mismatch=ignore_version_mismatch,
+            ignore_unrecognized_tag=ignore_unrecognized_tag,
+        )
+
+    generic_file = generic_io.get_file(io.BytesIO(asdf_extension.data), mode="rw")
+    af = asdf.open(
+        generic_file,
+        ignore_version_mismatch=ignore_version_mismatch,
+        ignore_unrecognized_tag=ignore_unrecognized_tag,
+        ignore_missing_extensions=ignore_missing_extensions,
+    )
+    # map hdulist to blocks here
+    _map_hdulist_to_arrays(hdulist, af)
+    return af
+
+
+def _map_hdulist_to_arrays(hdulist, af):
+    def callback(node, json_id):
+        if (
+                isinstance(node, NDArrayType) and
+                isinstance(node._source, str) and
+                node._source.startswith(_FITS_SOURCE_PREFIX)
+                ):
+            # read the array data from the hdulist
+            source = node._source
+            parts = re.match(
+                # All printable ASCII characters are allowed in EXTNAME
+                "((?P<name>[ -~]+),)?(?P<ver>[0-9]+)",
+                source[len(_FITS_SOURCE_PREFIX) :],
+            )
+            if parts is not None:
+                ver = int(parts.group("ver"))
+                if parts.group("name"):
+                    pair = (parts.group("name"), ver)
+                else:
+                    pair = ver
+            data = hdulist[pair].data
+            return data
+        return node
+    af.tree = treeutil.walk_and_modify(af.tree, callback)
 
 
 def from_fits_hdu(hdu, schema, cast_arrays=True):
